@@ -1,12 +1,15 @@
 import os
 import sys
+import logging
 import soco
 import soco.alarms
 import pprint
 import tabulate
+import datetime
 from collections import namedtuple
 
 from . import sonos
+from . import speaker_info
 
 pp = pprint.PrettyPrinter(width=120)
 
@@ -36,16 +39,70 @@ def convert_true_false(true_or_false, conversion="YesOrNo"):
         return "on" if true_or_false is True else "off"
 
 
+# Parameter checking decorators
+def zero_parameters(f):
+    def wrapper(*args, **kwargs):
+        if len(args[2]) != 0:
+            parameter_number_error(args[1], "no")
+            return False
+        else:
+            return f(*args, **kwargs)
+
+    return wrapper
+
+
+def one_parameter(f):
+    def wrapper(*args, **kwargs):
+        if len(args[2]) != 1:
+            parameter_number_error(args[1], "1")
+            return False
+        else:
+            return f(*args, **kwargs)
+
+    return wrapper
+
+
+def zero_or_one_parameter(f):
+    def wrapper(*args, **kwargs):
+        if len(args[2]) not in [0, 1]:
+            parameter_number_error(args[1], "0 or 1")
+            return False
+        else:
+            return f(*args, **kwargs)
+
+    return wrapper
+
+
+def one_or_two_parameters(f):
+    def wrapper(*args, **kwargs):
+        if len(args[2]) not in [1, 2]:
+            parameter_number_error(args[1], "1 or 2")
+            return False
+        else:
+            return f(*args, **kwargs)
+
+    return wrapper
+
+
+def two_parameters(f):
+    def wrapper(*args, **kwargs):
+        if len(args[2]) != 2:
+            parameter_number_error(args[1], "2")
+            return False
+        else:
+            return f(*args, **kwargs)
+
+    return wrapper
+
+
 # Action processing functions
+@zero_or_one_parameter
 def on_off_action(speaker, action, args, soco_function, use_local_speaker_list):
     """Method to deal with actions that have 'on|off semantics"""
-    np = len(args)
-    if np not in [0, 1]:
-        parameter_number_error(action, "0 or 1")
-        return False
     if action == "group_mute":
         speaker = speaker.group
         soco_function = "mute"
+    np = len(args)
     if np == 0:
         state = "on" if getattr(speaker, soco_function) else "off"
         print(state)
@@ -60,10 +117,8 @@ def on_off_action(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_parameters
 def no_args_no_output(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     if soco_function == "separate_stereo_pair" and float(soco.__version__) < 0.20:
         error_and_exit("Pairing operations require SoCo v0.20 or greater")
         return False
@@ -71,10 +126,8 @@ def no_args_no_output(speaker, action, args, soco_function, use_local_speaker_li
     return True
 
 
+@zero_parameters
 def no_args_one_output(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     result = getattr(speaker, soco_function)
     if callable(result):
         print(getattr(speaker, soco_function)())
@@ -83,10 +136,8 @@ def no_args_one_output(speaker, action, args, soco_function, use_local_speaker_l
     return True
 
 
+@zero_parameters
 def list_queue(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     queue = speaker.get_queue(max_items=1000)
     for i in range(len(queue)):
         try:
@@ -109,10 +160,8 @@ def list_queue(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_parameters
 def list_numbered_things(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     if soco_function in [
         "get_sonos_favorites",
         "get_favorite_radio_stations",
@@ -134,11 +183,9 @@ def list_numbered_things(speaker, action, args, soco_function, use_local_speaker
     return True
 
 
+@zero_or_one_parameter
 def volume_actions(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if np not in [0, 1]:
-        parameter_number_error(action, "0 or 1")
-        return False
     # Special case for ramp_to_volume
     if soco_function == "ramp_to_volume":
         if np == 1:
@@ -170,10 +217,8 @@ def volume_actions(speaker, action, args, soco_function, use_local_speaker_list)
     return True
 
 
+@one_parameter
 def relative_volume(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 1:
-        parameter_number_error(action, "1")
-        return False
     if soco_function == "group_relative_volume":
         speaker = speaker.group
     try:
@@ -189,22 +234,18 @@ def relative_volume(speaker, action, args, soco_function, use_local_speaker_list
     return True
 
 
+@zero_parameters
 def print_info(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     output = getattr(speaker, soco_function)()
     for item in sorted(output):
-        if item not in ["metadata", "uri"]:
+        if item not in ["metadata", "uri", "album_art"]:
             print("  {}: {}".format(item, output[item]))
     return True
 
 
+@zero_or_one_parameter
 def playback_mode(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if np not in [0, 1]:
-        parameter_number_error(action, "0 or 1")
-        return False
     possible_args = [
         "normal",
         "repeat_all",
@@ -222,19 +263,14 @@ def playback_mode(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_parameters
 def transport_state(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) == 0:
-        print(speaker.get_current_transport_info()["current_transport_state"])
-        return True
-    else:
-        parameter_number_error(action, "no")
-        return False
+    print(speaker.get_current_transport_info()["current_transport_state"])
+    return True
 
 
+@one_parameter
 def play_favourite(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 1:
-        parameter_number_error(action, "1")
-        return False
     favourite = args[0]
     fs = speaker.music_library.get_sonos_favorites()
     the_fav = None
@@ -273,10 +309,8 @@ def play_favourite(speaker, action, args, soco_function, use_local_speaker_list)
     return False
 
 
+@one_parameter
 def play_favourite_radio(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 1:
-        parameter_number_error(action, "1")
-        return False
     favourite = args[0]
     fs = speaker.music_library.get_favorite_radio_stations()
     the_fav = None
@@ -315,6 +349,7 @@ def play_favourite_radio(speaker, action, args, soco_function, use_local_speaker
     return False
 
 
+@one_or_two_parameters
 def play_uri(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
     if np not in [1, 2]:
@@ -331,11 +366,9 @@ def play_uri(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_or_one_parameter
 def sleep_timer(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if np not in [0, 1]:
-        parameter_number_error(action, "0 or 1")
-        return False
     if np == 0:
         st = speaker.get_sleep_timer()
         if st:
@@ -343,21 +376,48 @@ def sleep_timer(speaker, action, args, soco_function, use_local_speaker_list):
         else:
             print(0)
     elif np == 1:
-        try:
-            t = int(args[0])
-            if not 0 <= t <= 86399:
-                raise Exception
-        except:
-            parameter_type_error(action, "integer > 0")
-            return False
-        speaker.set_sleep_timer(int(args[0]))
+        if args[0].lower() in ["off", "cancel"]:
+            logging.info("Cancelling sleep timer")
+            speaker.set_sleep_timer(None)
+        else:
+            duration = sonos.convert_to_seconds(args[0])
+            if duration is None:
+                parameter_type_error(
+                    action,
+                    "number of hours, seconds or minutes + 'h/m/s', or off|cancel",
+                )
+                return False
+            if 0 <= duration <= 86399:
+                logging.info("Setting sleep timer to {}s".format(duration))
+                speaker.set_sleep_timer(duration)
+            else:
+                parameter_type_error(action, "maximum duration is 23.999hrs")
+                return False
     return True
 
 
-def group_or_pair(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 1:
-        parameter_number_error(action, "1")
+@one_parameter
+def sleep_at(speaker, action, args, soco_function, use_local_speaker_list):
+    try:
+        off_time = datetime.time.fromisoformat(args[0])
+        now_time = datetime.datetime.now().time()
+        delta_off = datetime.timedelta(hours=off_time.hour, minutes=off_time.minute, seconds=off_time.second)
+        delta_now = datetime.timedelta(hours=now_time.hour, minutes=now_time.minute, seconds=now_time.second)
+        duration = int((delta_off - delta_now).total_seconds())
+    except ValueError:
+        parameter_type_error(action, "a time in 24hr 'HH:MM' or 'HH:MM:SS' format")
         return False
+    if 0 <= duration <= 86399:
+        logging.info("Setting sleep timer to {}s".format(duration))
+        speaker.set_sleep_timer(duration)
+    else:
+        parameter_type_error(action, "maximum duration is 23.999hrs")
+        return False
+    return True
+
+
+@one_parameter
+def group_or_pair(speaker, action, args, soco_function, use_local_speaker_list):
     if soco_function == "create_stereo_pair" and float(soco.__version__) < 0.20:
         error_and_exit("Pairing operations require SoCo v0.20 or greater")
         return False
@@ -366,10 +426,8 @@ def group_or_pair(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_parameters
 def operate_on_all(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     zones = speaker.all_zones
     for zone in zones:
         if zone.is_visible:
@@ -383,21 +441,17 @@ def operate_on_all(speaker, action, args, soco_function, use_local_speaker_list)
     return True
 
 
+@zero_parameters
 def zones(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     zones = speaker.all_zones if "all" in action else speaker.visible_zones
     for zone in zones:
         print("{} ({})".format(zone.player_name, zone.ip_address))
     return True
 
 
+@zero_or_one_parameter
 def play_from_queue(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if np not in [0, 1]:
-        parameter_number_error(action, "0 or 1")
-        return False
     if np == 0:
         speaker.play_from_queue(0)
     elif np == 1:
@@ -414,10 +468,8 @@ def play_from_queue(speaker, action, args, soco_function, use_local_speaker_list
     return True
 
 
+@one_parameter
 def remove_from_queue(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 1:
-        parameter_number_error(action, "1")
-        return False
     try:
         index = int(args[0])
     except:
@@ -432,18 +484,14 @@ def remove_from_queue(speaker, action, args, soco_function, use_local_speaker_li
     return True
 
 
+@one_parameter
 def save_queue(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 1:
-        parameter_number_error(action, "1")
-        return False
     speaker.create_sonos_playlist_from_queue(args[0])
     return True
 
 
+@one_parameter
 def seek(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 1:
-        parameter_number_error(action, "1")
-        return False
     try:
         speaker.seek(args[0])
     except:
@@ -452,10 +500,8 @@ def seek(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@one_parameter
 def playlist_operations(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 1:
-        parameter_number_error(action, "1")
-        return False
     name = args[0]
     if soco_function == "create_sonos_playlist":
         getattr(speaker, soco_function)(name)
@@ -479,10 +525,8 @@ def playlist_operations(speaker, action, args, soco_function, use_local_speaker_
     return False
 
 
+@two_parameters
 def remove_from_playlist(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 2:
-        parameter_number_error(action, "2")
-        return False
     name = args[0]
     try:
         track_number = int(args[1])
@@ -505,11 +549,9 @@ def remove_from_playlist(speaker, action, args, soco_function, use_local_speaker
     return False
 
 
+@zero_or_one_parameter
 def line_in(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if np not in [0, 1]:
-        parameter_number_error(action, "0 or 1")
-        return False
     if np == 0:
         state = "on" if speaker.is_playing_line_in else "off"
         print(state)
@@ -525,11 +567,9 @@ def line_in(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_or_one_parameter
 def eq(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if np not in [0, 1]:
-        parameter_number_error(action, "0 or 1")
-        return False
     if np == 0:
         print(getattr(speaker, soco_function))
     elif np == 1:
@@ -546,11 +586,9 @@ def eq(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_or_one_parameter
 def balance(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if np not in [0, 1]:
-        parameter_number_error(action, "0 or 1")
-        return False
     if np == 0:
         left, right = getattr(speaker, soco_function)
         # Convert to something more intelligible than a 2-tuple
@@ -576,18 +614,14 @@ def balance(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_parameters
 def reindex(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     speaker.music_library.start_library_update()
     return True
 
 
+@zero_parameters
 def info(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     info = speaker.get_speaker_info()
     model = info["model_name"].lower()
     if not ("boost" in model or "bridge" in model):
@@ -619,10 +653,8 @@ def info(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_parameters
 def groups(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     for group in speaker.all_groups:
         if group.coordinator.is_visible:
             print("[{}] : ".format(group.short_label), end="")
@@ -634,10 +666,8 @@ def groups(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_parameters
 def list_alarms(speaker, action, args, soco_function, use_local_speaker_list):
-    if len(args) != 0:
-        parameter_number_error(action, "no")
-        return False
     alarms = soco.alarms.get_alarms(speaker)
     if not alarms:
         return True
@@ -680,7 +710,23 @@ def list_alarms(speaker, action, args, soco_function, use_local_speaker_list):
         "Recurrence",
         "Include Grouped",
     ]
-    print(tabulate.tabulate(sorted(details), headers))
+    print(tabulate.tabulate(sorted(details), headers, numalign="center"))
+    return True
+
+
+@zero_parameters
+def list_libraries(speaker, action, args, soco_function, use_local_speaker_list):
+    shares = speaker.music_library.list_library_shares()
+    index = 0
+    for share in sorted(shares):
+        index += 1
+        print("{:2d}: {}".format(index, share))
+    return True
+
+
+@zero_parameters
+def system_info(speaker, action, args, soco_function, use_local_speaker_list):
+    speaker_info.print_speaker_table(speaker)
     return True
 
 
@@ -764,6 +810,7 @@ actions = {
     "group": SonosFunction(group_or_pair, "join"),
     "g": SonosFunction(group_or_pair, "join"),
     "ungroup": SonosFunction(no_args_no_output, "unjoin"),
+    "ug": SonosFunction(no_args_no_output, "unjoin"),
     "u": SonosFunction(no_args_no_output, "unjoin"),
     "party_mode": SonosFunction(no_args_no_output, "partymode"),
     "party": SonosFunction(no_args_no_output, "partymode"),
@@ -824,4 +871,8 @@ actions = {
     "artists": SonosFunction(list_numbered_things, "get_artists"),
     "tracks": SonosFunction(list_numbered_things, "get_tracks"),
     "alarms": SonosFunction(list_alarms, "get_alarms"),
+    "libraries": SonosFunction(list_libraries, "list_library_shares"),
+    "shares": SonosFunction(list_libraries, "list_library_shares"),
+    "sysinfo": SonosFunction(system_info, ""),
+    "sleep_at": SonosFunction(sleep_at, "")
 }
